//
// Created by Иван_Архипов on 31.10.2017.
//

#include "DatFile.h"

#include "BinaryData.h"
#include "DatException.h"
#include "SubDirectory.h"
#include "Subfile.h"
#include "SubfileData.h"

#include "EasyLogging++/easylogging++.h"
#define ELPP_FEATURE_CRASH_LOG
INITIALIZE_EASYLOGGINGPP
#include <locale>

#ifdef WIN32
#define fseek _fseeki64
#define ftell _ftelli64
#endif

extern "C++"
{
namespace LOTRO_DAT {

    DatFile::DatFile() {
        dat_state_ = CLOSED;
        root_directory_ = nullptr;
        file_handler_ = nullptr;

        el::Configurations defaultConf;
        defaultConf.setToDefault();
        defaultConf.setGlobally(el::ConfigurationType::Format, "%datetime %level %fbase (line %line) : %msg (function: %func)");
        defaultConf.setGlobally(el::ConfigurationType::ToFile, "true");
        defaultConf.setGlobally(el::ConfigurationType::Filename, "dat_library.log");
        defaultConf.setGlobally(el::ConfigurationType::ToStandardOutput, "false");
        defaultConf.setGlobally(el::ConfigurationType::PerformanceTracking, "true");
        defaultConf.setGlobally(el::ConfigurationType::MaxLogFileSize, "20971520"); // 20MB
        defaultConf.setGlobally(el::ConfigurationType::LogFlushThreshold, "1"); // Flush after every one log
        el::Loggers::reconfigureAllLoggers(defaultConf);

        LOG(INFO) << "Starting new DatFile class instance";
    }

    DAT_RESULT DatFile::InitDatFile(const std::string &filename, int dat_id) {
        if (dat_state_ != CLOSED && filename == filename_) {
            LOG(DEBUG) << "Trying to reopen the same file: " << filename << ". Doing nothing.";
            return SUCCESS;
        }

        if (dat_state_ != CLOSED && filename != filename_) {
            LOG(DEBUG) << "DatFile wasn't closed yet. Closing in order to reopen.";
            if (CloseDatFile() != SUCCESS) {
                LOG(ERROR) << "Unable to perform CloseDatFile()! Aborting initialization!";
                return FAILED;
            }
        }

        dat_id_ = dat_id;
        dat_state_ = CLOSED;
        current_locale_ = ORIGINAL;
        root_directory_ = nullptr;
        file_handler_ = nullptr;
        filename_ = "none";

        DAT_RESULT result;

        result = OpenDatFile(filename.c_str());
        if (result != SUCCESS) {
            LOG(ERROR) << "Unable to perform opening file. Aborting.";
            CloseDatFile();
            return result;
        }

        result = ReadSuperBlock();
        if (result != SUCCESS) {
            LOG(ERROR) << "Unable to read super block. Aborting.";
            CloseDatFile();
            return result;
        }

        result = MakeDirectories();
        if (result != SUCCESS) {
            LOG(ERROR) << "Unable to make directories. Aborting.";
            CloseDatFile();
            return result;
        }

        result = MakeDictionary();
        if (result != SUCCESS) {
            LOG(ERROR) << "Unable to make dictionary. Aborting.";
            CloseDatFile();
            return result;
        }

        result = InitLocales();
        if (result != SUCCESS) {
            LOG(ERROR) << "Unable to initialize locales. Aborting.";
            CloseDatFile();
            return result;
        }

        LOG(INFO) << "File " << filename << " opened successfully!";
        filename_ = filename;
        dat_state_ = READY;
        return SUCCESS;
    }

    DAT_STATE DatFile::DatFileState() const {
        return dat_state_;
    }

    DatFile::~DatFile() {
        CloseDatFile();
    }

    /// Extracts file with file_id.
    /// If path is undefined then it will be recognised as current working directory
    /// Output file path consists of "path + file_id + file_extension";
    /// NOTICE: The directory, mentioned in "std::string path" variable SHOULD BE ALREADY CREATED;
    /// Otherwise DatException() will be thrown.
    /// Returns true, if file was successfully extracted;
    /// Throws DatException() if undefined behaviour happened

    DAT_RESULT DatFile::ExtractFile(long long file_id, const std::string &path) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot perform extraction.";
            return INCORRECT_STATE_ERROR;
        }
        BinaryData file_data = GetFileData(dictionary_[file_id], 8);

        if (file_data.size() == 0) {
            LOG(ERROR) << "File data is empty. Aborting extraction.";
            return NO_FILE_ERROR;
        }

        SubfileData export_data = dictionary_[file_id]->PrepareForExport(file_data);

        if (export_data.Empty()) {
            LOG(ERROR) << "Export data is empty. Aborting extraction.";
            return NO_FILE_ERROR;
        }

        if (export_data.binary_data.WriteToFile(path + export_data.options["ext"].as<std::string>()) != SUCCESS) {
            LOG(ERROR) << "Cannot write to file.";
            return WRITE_TO_FILE_ERROR;
        }
        return SUCCESS;
    }

    /// Extracts file with file_id to database "db".
    /// DATABASE SHOULD BE ALREADY CREATED; Otherwise DatException will be called.
    /// NOTICE: The directory, mentioned in "std::string path" variable SHOULD BE ALREADY CREATED;
    /// Otherwise DatException() will be thrown.
    /// Returns true, if file was successfully extracted;
    /// Throws DatException() if undefined behaviour happened

    DAT_RESULT DatFile::ExtractFile(long long file_id, Database *db) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot perform extraction.";
            return INCORRECT_STATE_ERROR;
        }

        BinaryData file_data;

        try {
            file_data = GetFileData(dictionary_[file_id], 8);
        } catch (std::exception &e) {
            LOG(ERROR) << "Caught" << e.what() << " exception.";
            return FAILED;
        }

        SubfileData export_data;

        try {
            export_data = dictionary_[file_id]->PrepareForExport(file_data);
            export_data.options["did"] = dat_id_;
        } catch (std::exception &e) {
            LOG(ERROR) << "Caught" << e.what() << " exception.";
            return FAILED;
        }

        if (export_data == SubfileData()) {
            LOG(WARNING) << "file with id " << dictionary_[file_id]->file_id() << " is empty. Passing it.";
            return SUCCESS;
        }

        try {
            db->PushFile(export_data);
        } catch (std::exception &e) {
            LOG(ERROR) << "Caught " << e.what() << " exception.";
            return FAILED;
        }
        return SUCCESS;
    }

    /// Extracts all files with specific type to "path + type + file_id + file_part + extension" files;
    /// If path is undefined then it will be recognised as current working directory
    /// NOTICE: The directory, mentioned in "std::string path" variable SHOULD BE ALREADY CREATED;
    /// Otherwise DatException() will be thrown.
    /// Returns number of successfully extracted files
    /// Throws DatException() if undefined behaviour happened

    int DatFile::ExtractAllFilesByType(FILE_TYPE type, std::string path) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot perform extraction.";
            return INCORRECT_STATE_ERROR;
        }

        int success = 0;
        for (auto i : dictionary_) {
            FILE_TYPE file_type = i.second->FileType();
            if (file_type  == type) {
                success += (ExtractFile(i.second->file_id(), (path + std::to_string(i.second->file_id()))) == SUCCESS ? 1 : 0);
            }
        }
        return success;
    }

    /// Extracts all files with specific type to database "db";
    /// DATABASE SHOULD BE ALREADY CREATED; Otherwise DatException will be called.
    /// Returns number of successfully extracted files
    /// Throws DatException() if undefined behaviour happened

    int DatFile::ExtractAllFilesByType(FILE_TYPE type, Database *db) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot perform extraction.";
            return INCORRECT_STATE_ERROR;
        }

        int success = 0;
        for (auto i : dictionary_) {
            FILE_TYPE file_type = i.second->FileType();
            if (file_type  == type) {
                success += (ExtractFile(i.second->file_id(), db) == SUCCESS ? 1 : 0);
            }
        }
        return success;
    }

    // TODO: Write description and make asserts
    DAT_RESULT DatFile::PatchFile(const char *filename, YAML::Node options) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot patch.";
            return INCORRECT_STATE_ERROR;
        }

        if (options["did"].IsDefined() && options["did"].as<int>() != dat_id_)
            return INCORRECT_DAT_ID;

        BinaryData data;
        data.ReadFromFile(filename);

        auto file_id = options["fid"].as<long long>();

        if (dictionary_[file_id] == nullptr) {
            LOG(ERROR) <<  "Cannot patch file - there is no file in dictionary with file_id = " << file_id;
            return NO_FILE_ERROR;
        }

        BinaryData old_data = GetFileData(dictionary_[file_id]);
        data = dictionary_[file_id]->MakeForImport(old_data, SubfileData(data, u"", options));

        try {
            DAT_RESULT result = ApplyFilePatch(dictionary_[file_id], data);
            if (result != SUCCESS)
                return result;
        } catch (std::exception &e) {
            LOG(ERROR) << "Caught " << e.what() <<" exception.";
            return FAILED;
        }
        return SUCCESS;
    }

    // TODO: Write description and make asserts
    DAT_RESULT DatFile::PatchFile(const SubfileData &data, bool rewrite_original) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot patch.";
            return INCORRECT_STATE_ERROR;
        }

        auto file_id = data.options["fid"].as<long long>();
        Subfile *file = dictionary_[file_id];

        if (file == nullptr) {
            LOG(ERROR) << "Cannot patch file - there is no file in dictionary with file_id = " << file_id;
            return NO_FILE_ERROR;
        }

        // If file has inactive category, then we should set it to patched state in order to commit patch and
        // then in ApplyFilePatch(), if new category is still inactive, return dictionary to its original state;

        if (inactive_categories.count(file->category) != 0 && patch_dict_.count(file_id) != 0) {
            dictionary_[file_id]->file_offset_ = patch_dict_[file_id]->file_offset_;
            dictionary_[file_id]->file_size_ = patch_dict_[file_id]->file_size_;
            dictionary_[file_id]->block_size_ = patch_dict_[file_id]->block_size_;
            dictionary_[file_id]->timestamp_ = patch_dict_[file_id]->timestamp_;
            dictionary_[file_id]->version_ = patch_dict_[file_id]->version_;
        }

        if (data.options["cat"].IsDefined()) {
            file->category = data.options["cat"].as<long long>();
        } else {
            file->category = 1;
        }

        BinaryData old_data = GetFileData(file);
        BinaryData patch_data = file->MakeForImport(old_data, data);
        try {
            DAT_RESULT result = ApplyFilePatch(file, patch_data, rewrite_original);
            if (result != SUCCESS)
                return result;
        } catch (std::exception &e) {
            LOG(ERROR) << "Caught " << e.what() << " exception";
            return FAILED;
        }
        return SUCCESS;
    }

    // TODO: Write description
    DAT_RESULT DatFile::PatchAllDatabase(Database *db) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot patch.";
            return INCORRECT_STATE_ERROR;
        }

        SubfileData data;
        data = db->GetNextFile();

        while (!data.Empty()) {
            DAT_RESULT result = PatchFile(data);
            if (result != SUCCESS)
                LOG(ERROR) << "Cannot patch file" << data.options["fid"].as<long long>() << " continuing";
                data = db->GetNextFile();
        }
        DAT_RESULT result = CommitChanges();
        if (result != SUCCESS)
            return result;

        return SUCCESS;
    }

    /// DatFile::WriteUnorderedDictionary(...);
    /// Prints list of all found files with some information about them to file.
    /// Gets std::string path - path to directory, where the file will be written with name "dict.txt"

    DAT_RESULT DatFile::WriteUnorderedDictionary(std::string path) const {
        FILE *f = nullptr;
        fopen_s(&f, (path + "dict.txt").c_str(), "w");

        if (f == nullptr) {
            LOG(ERROR) << "Cannot open file " << path + "dict.txt";
            return WRITE_TO_FILE_ERROR;
        }

        fprintf(f, "file_id offset size size2 extension\n");
        for (auto i : dictionary_) {
            fprintf(f, "%lld %lld %lld %lld %s\n", i.second->file_id(), i.second->file_offset(), i.second->file_size(),
                    i.second->block_size(), i.second->Extension().c_str());
        }
        fclose(f);
        return SUCCESS;
    }

    /// DatFile::files_number();
    /// Returns amount of files, found in dictionaries of DatFile. Some if them may be empty or erased.

    long long DatFile::files_number() const {
        return dictionary_.size();
    }

    /// DatFile::GetFileData()
    /// Returns BinaryData, which contains of subfile data, made from parts of file in DatFile
    // TODO: ASSERTS
    BinaryData DatFile::GetFileData(const Subfile *file, long long int offset) {
        BinaryData mfile_id(20);
        ReadData(mfile_id, 20, file->file_offset() + 8);
        if (!mfile_id.CheckCompression() && file->file_id() != mfile_id.ToNumber<4>(0))
            throw DatException("Bad DatFile::GetFileData() - file_id in Subfile doesn't match to file_id in DatFile.", READ_EXCEPTION);

        BinaryData data((unsigned)(file->file_size() + (8 - offset)));
        if (file->block_size() >= file->file_size() + 8) {
            ReadData(data, file->file_size() + (8 - offset), file->file_offset() + offset);
            return data;
        }

        BinaryData fragments_count(4);
        ReadData(fragments_count, 4, file->file_offset());

        long long fragments_number = fragments_count.ToNumber<4>(0);

        long long current_block_size = file->block_size() - offset - 8 * fragments_number;

        ReadData(data, current_block_size , file->file_offset() + offset);

        BinaryData FragmentsDictionary(8 * unsigned(fragments_number));
        ReadData(FragmentsDictionary, 8 * unsigned(fragments_number), file->file_offset() + file->block_size() - 8 * fragments_number);


        for (long long i = 0; i < fragments_number; i++) {
            long long fragment_size = FragmentsDictionary.ToNumber<4>(8 * i);
            long long fragment_offset = FragmentsDictionary.ToNumber<4>(8 * i + 4);
            ReadData(data, std::min(fragment_size, file->file_size() - current_block_size), fragment_offset, current_block_size );
            current_block_size += fragment_size;
        }

        return data;
    }

    /// DatFile special functions for opening and reading/writing raw data.
    /// Shouldn't be used by any external classes except Subfile and Subdirectory.

    DAT_RESULT DatFile::OpenDatFile(const char *dat_name) {
        if (dat_state_ != CLOSED) {
            CloseDatFile();
        }

        fopen_s(&file_handler_, dat_name, "r+b");

        if (file_handler_ == nullptr) {
            LOG(ERROR) << "Unable to open file " << dat_name;
            return NO_FILE_ERROR;
        }

        fseek(file_handler_, 0, SEEK_END);
        file_size_ = ftell(file_handler_);
        fseek(file_handler_, 0, SEEK_SET);

        dat_state_ = SUCCESS_OPENED;
        return SUCCESS;
    }

    DAT_RESULT DatFile::ReadSuperBlock() {
        if (dat_state_ != SUCCESS_OPENED) {
            LOG(ERROR) << "Dat state isn't SUCCESS_OPENED. Cannot perform extraction.";
            return INCORRECT_STATE_ERROR;
        }

        BinaryData data(1024);
        ReadData(data, 1024);

        constant1_ = data.ToNumber<4>(0x100);
        constant2_ = data.ToNumber<4>(0x140);
        version1_ = data.ToNumber<4>(0x14C);
        version2_ = data.ToNumber<4>(0x150);
        fragmentation_journal_offset_ = data.ToNumber<4>(0x154);
        root_directory_offset_ = data.ToNumber<4>(0x160);
        auto size1 = data.ToNumber<4>(0x148);

        if (constant1_ != 0x4C5000) {
            LOG(ERROR) << "variable at position 0x100 is not equal to .dat file constant!";
            return INCORRECT_SUPERBLOCK_ERROR;
        }
        if (constant2_ != 0x5442) {
            LOG(ERROR) << "variable at position 0x140 is not equal to .dat file constant!";
            return INCORRECT_SUPERBLOCK_ERROR;
        }

        if (file_size_ != size1) {
            LOG(ERROR) << "variable at 0x148 position is not equal to .dat file size!";
            return INCORRECT_SUPERBLOCK_ERROR;
        }

        dat_state_ = SUCCESS_SUPERBLOCK;
        return SUCCESS;
    }

    DAT_RESULT DatFile::MakeDirectories() {
        if (dat_state_ != SUCCESS_SUPERBLOCK) {
            LOG(ERROR) << "Dat state isn't SUCCESS_SUPERBLOCK. Cannot make directories.";
            return INCORRECT_STATE_ERROR;
        }

        root_directory_ = new SubDirectory((unsigned) root_directory_offset_, this);
        dat_state_ = SUCCESS_DIRECTORIES;
        return SUCCESS;
    }

    DAT_RESULT DatFile::MakeDictionary() {
        if (dat_state_ != SUCCESS_DIRECTORIES) {
            LOG(ERROR) << "Dat state isn't SUCCESS_DIRECTORIES. Cannot make directories.";
            return INCORRECT_STATE_ERROR;
        }

        if (root_directory_ == nullptr) {
            LOG(ERROR) << "root_directory is nullptr!!";
            return INIT_ERROR;
        }

        root_directory_->MakeDictionary(dictionary_);
        dat_state_ = SUCCESS_DICTIONARY;
        return SUCCESS;
    }

    DAT_RESULT DatFile::ReadData(BinaryData &data, long long size, long long offset, long long data_offset) {
        if (dat_state_ == CLOSED) {
            LOG(ERROR) << "Dat state is CLOSED. Cannot read data.";
            return INCORRECT_STATE_ERROR;
        }

        if (data_offset + size > data.size()) {
            LOG(ERROR) << "Trying to read more than BinaryData size: Reading " << size << " bytes from " << offset << " position.";
            return DAT_READ_ERROR;
        }

        if (offset + size > file_size_) {
            LOG(ERROR) << "Trying to read more than DatFile size elapsed: Reading " << size << " bytes from " << offset << " position.";
            return DAT_READ_ERROR;
        }

        fseek(file_handler_, offset, SEEK_SET);
        fread(data.data() + data_offset, unsigned(size), 1, file_handler_);
        return SUCCESS;
    }

    DAT_RESULT DatFile::WriteData(const BinaryData &data, long long size, long long offset, long long data_offset) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot write data.";
            return INCORRECT_STATE_ERROR;
        }

       fseek(file_handler_, offset, SEEK_SET);
        if (data_offset + size > data.size())
            throw DatException("Bad DatFile::WriteData - trying to write more than BinaryData size", WRITE_EXCEPTION);

        fwrite(data.data() + data_offset, unsigned(size), 1, file_handler_);
        return SUCCESS;
    }

    /// Special functions used by patch process.
    /// Shouldn't be used by any external class.
    DAT_RESULT DatFile::ApplyFilePatch(Subfile *file, const BinaryData &data, bool rewrite_original) {
        auto file_id = file->file_id();
        if (patched_list.count(file_id) != 0) {
            LOG(WARNING) << "Warning: DatFile::ApplyFilePatch - found 2 files in patch with the same file_id = "
                         << file->file_id() << " Passing last...";
            return DUBLICATE_PATCH_FILES_ERROR;
        }

        if (current_locale() != PATCHED && !rewrite_original) {
            LOG(INFO) << "Changing locale to PATCHED(RU) in order to patch file";
            SetLocale(PATCHED);
        }

        if (current_locale() == PATCHED && rewrite_original && patch_dict_.count(file_id) != 0) {
            LOG(INFO) << "Changing locale to ORIGINAL in order to patch original version of file, which has both "
                      << "original and patched versions.";
            SetLocale(ORIGINAL);
        }

        dat_state_ = UPDATED;

        if (orig_dict_.count(file_id) == 0 && !rewrite_original) {
            orig_dict_[file_id] = new Subfile(this, file->MakeHeaderData());
        }

        auto journal = GetFragmentationJournal();

        file->file_size_ = data.size() - 8;

        if ((patch_dict_.count(file_id) == 0 && !rewrite_original) || data.size() > file->block_size()) {
            if (journal[0].second != file_size_) {
                journal[0].second = file_size_;
            }

            file->file_offset_ = journal[0].second;
            file->block_size_ = std::max(data.size(), 256u);

            journal[0].second += data.size();

            BinaryData nulls(data.size());
            WriteData(nulls, nulls.size(), file_size_);

            this->file_size_ += data.size();
        }

        BinaryData fragments_count(4);
        fragments_count = BinaryData::FromNumber<4>(0);

        BinaryData file_data = fragments_count + data.CutData(4);

        if (file_id != file_data.ToNumber<4>(8)) {
            LOG(ERROR) << "Created data's file_id doesn't match to original! Patch wasn't written to .dat file";
            return INCORRECT_PATCH_FILE;
        }

        WriteData(file_data, file_data.size(), file->file_offset());

        patched_list.insert(file_id);

        if (!rewrite_original) {
            patch_dict_.erase(file_id); // Удалили старое значение в русском словаре
            patch_dict_[file_id] = new Subfile(this, file->MakeHeaderData()); // Создали новое значение
        }

        // If category is forbidden, then return file header data to original state
        if (inactive_categories.count(file->category) != 0) {
            dictionary_[file_id]->file_offset_ = orig_dict_[file_id]->file_offset_;
            dictionary_[file_id]->file_size_ = orig_dict_[file_id]->file_size_;
            dictionary_[file_id]->block_size_ = orig_dict_[file_id]->block_size_;
            dictionary_[file_id]->timestamp_ = orig_dict_[file_id]->timestamp_;
            dictionary_[file_id]->version_ = orig_dict_[file_id]->version_;
        }

        if (orig_dict_.count(file_id) != 0)
            orig_dict_[file_id]->category = file->category;
        if (patch_dict_.count(file_id) != 0)
            patch_dict_[file_id]->category = file->category;

        UpdateFragmentationJournal(journal);
        return SUCCESS;
    }

    DAT_RESULT DatFile::UpdateSubdirectories() {
        // TODO: asserts
        root_directory_->UpdateDirectories(patched_list, dictionary_);
        return SUCCESS;
    }

    std::vector<std::pair<long long, long long> > DatFile::GetFragmentationJournal() {
        BinaryData data(8);
        ReadData(data, 8, fragmentation_journal_offset_ + 8);
        std::vector<std::pair<long long, long long> > result;
        result.emplace_back(std::make_pair(data.ToNumber<4>(0), data.ToNumber<4>(4)));
        return result;
    }

    DAT_RESULT DatFile::UpdateHeader() {
        WriteData(BinaryData::FromNumber<4>(constant1_), 4, 0x100);
        WriteData(BinaryData::FromNumber<4>(constant2_), 4, 0x140);
        WriteData(BinaryData::FromNumber<4>(file_size_), 4, 0x148);
        WriteData(BinaryData::FromNumber<4>(version1_), 4, 0x14C);
        WriteData(BinaryData::FromNumber<4>(version2_), 4, 0x150);
        WriteData(BinaryData::FromNumber<4>(fragmentation_journal_offset_), 4, 0x154);
        WriteData(BinaryData::FromNumber<4>(root_directory_offset_), 4, 0x160);
        return SUCCESS;
    }

    DAT_RESULT DatFile::UpdateFragmentationJournal(const std::vector<std::pair<long long, long long> > &journal) {
        for (unsigned i = 0; i < journal.size(); i++) {
            long long size = journal[i].first;
            long long offset = journal[i].second;

            WriteData(BinaryData::FromNumber<4>(size), 4, fragmentation_journal_offset_ + 8 * (i + 1));
            WriteData(BinaryData::FromNumber<4>(offset), 4, fragmentation_journal_offset_ + 8 * (i + 1) + 4);
        }
        return SUCCESS;
    }

    DAT_RESULT DatFile::CommitChanges() {
        if (dat_state_ != UPDATED) {
            LOG(DEBUG) << "Commiting changes to file with state != UPDATED. Nothing to do";
            return SUCCESS;
        }

        LOG(DEBUG) << "There are some updated files. Rewriting dictionary...";
        LOG(DEBUG) << "Updating locales...";
        CommitLocales();

        auto journal = GetFragmentationJournal();
        UpdateFragmentationJournal(journal);
        LOG(DEBUG) << "Updated fragmentation journal...";

        UpdateHeader();
        LOG(DEBUG) << "Updated header...";
        UpdateSubdirectories();
        LOG(DEBUG) << "Updated subdirectories...";
        LOG(DEBUG) << "Changed " << patched_list.size() << " files...";

        patched_list.clear();
        dat_state_ = READY;
        LOG(DEBUG) << "Done Commiting changes!";
        return SUCCESS;
    }

    DAT_RESULT DatFile::CloseDatFile() {
        if (dat_state_ == CLOSED) {
            LOG(INFO) << "DatFile is already closed. Nothing to do";
            return SUCCESS;
        }

        CommitChanges();

        orig_dict_.clear();
        patched_list.clear();
        pending_patch_.clear();

        current_locale_ = ORIGINAL;
        filename_ = "none";

        if (file_handler_ != nullptr)
            fclose(file_handler_);
        delete file_handler_;

        delete root_directory_;

        patched_list.clear();
        dictionary_.clear();

        dat_state_ = CLOSED;

        return SUCCESS;
    }

    // LOCALE MANAGING SECTION

    DAT_RESULT DatFile::InitLocales() {
        LOG(DEBUG) << "Initialising locales...";
        BinaryData dicts_data;

        if (dictionary_.count(2013266257) != 0)
            dicts_data = GetFileData(dictionary_[2013266257]);

        if (dicts_data.size() < 29) {
            LOG(WARNING) << "locales file is empty.. Initialising locale dicts as empty";
            LOG(INFO) << "Could't find locales file... Continuing without them";
            return SUCCESS;
        }

        BinaryData hi_data = dicts_data.CutData(14, 29) + BinaryData("\0", 1);
        std::string hi = std::string((char*)(hi_data.data()));
        LOG(DEBUG) << "hi info is " << hi;

        if (hi != "Hi from Gi1dor!") {
            LOG(WARNING) << "Didn't receive 'hi' from Gi1dor... Initialising locale dicts as empty";
            LOG(INFO) << "Could't init locales' file... Continuing without them";
            return SUCCESS;
        }

        int offset = 29;
        BinaryData current_locale_data = dicts_data.CutData(offset, offset + 4) + BinaryData("\0", 1);
        std::string locale((char*)(current_locale_data.data()));
        offset += 4;
        LOG(DEBUG) << "current locale:" << locale;

        if (locale != "PATC" && locale != "ORIG") {
            LOG(WARNING) << "Incorrect locale... Initialising locale dicts as empty";
            LOG(INFO) << "Could't recognize locale... Continuing without locales";
            return SUCCESS;
        }
        current_locale_ = (locale == "PATC" ? PATCHED : ORIGINAL);

        // 14 bytes for old data
        // 15 bytes for "Hi from Gi1dor"
        // 4 bytes for LOCALE
        // 4 bytes for orig_dict.size()
        // (32 + 4) * orig_dict.size() bytes for orig_dict data
        // 4 bytes for patch_dict.size()
        // (32 + 4) * patch_dict.size() bytes for patch_dict data
        // 4 bytes for inactive_categories dict
        // 4 * inactive_categories.size() bytes for inactive_categories data

        size_t orig_dict_size = size_t(dicts_data.CutData(offset, offset + 4).ToNumber<4>(0));
        offset += 4;
        for (size_t i = 0; i < orig_dict_size; i++) {
            auto file = new Subfile(this, dicts_data.CutData(offset, offset + 32));
            orig_dict_[file->file_id()] = file;
            offset += 32;
            orig_dict_[file->file_id()]->category = dicts_data.ToNumber<4>(offset);
            offset += 4;

            if (orig_dict_[file->file_id()]->category == 0)
                LOG(DEBUG) << "file category is undefined (0)!";
        }

        size_t patch_dict_size = size_t(dicts_data.CutData(offset, offset + 4).ToNumber<4>(0));
        offset += 4;
        for (size_t i = 0; i < patch_dict_size; i++) {
            auto file = new Subfile(this, dicts_data.CutData(offset, offset + 32));
            patch_dict_[file->file_id()] = file;
            offset += 32;
            patch_dict_[file->file_id()]->category = dicts_data.ToNumber<4>(offset);
            offset += 4;
            if (patch_dict_[file->file_id()]->category == 0)
                LOG(DEBUG) << "file category is undefined (0)!";

        }

        size_t active_patches_dict_size = size_t(dicts_data.CutData(offset, offset + 4).ToNumber<4>(0));
        offset += 4;
        for (size_t i = 0; i < active_patches_dict_size; i++) {
            inactive_categories.insert(dicts_data.ToNumber<4>(offset));
            offset += 4;
        }

        LOG(INFO) << "There are " << patch_dict_.size() << " files in patch locale dictionary";
        LOG(INFO) << "There are " << orig_dict_.size() << " files in original locale dictionary";
        std::string inactive_cat_s;
        for (auto i : inactive_categories) {
            inactive_cat_s += std::to_string(i) + " ";
        }
        LOG(INFO) << "Unactive patches now: " << inactive_cat_s;
        return SUCCESS;
    }

    std::map<long long, Subfile *> *DatFile::GetLocaleDictReference(LOCALE locale) {
        switch (locale) {
            case PATCHED:
                return &patch_dict_;
            case ORIGINAL:
                return &orig_dict_;
            default:
                LOG(ERROR) << "Unknown locale! Returning original";
                return &orig_dict_;
        }
    }

    DAT_RESULT DatFile::SetLocale(LOCALE locale) {
        if (dat_state_ < READY) {
            LOG(ERROR) << "Dat state isn't READY. Cannot set locale.";
            return INCORRECT_STATE_ERROR;
        }

        if (current_locale_ == locale) {
            return SUCCESS;
        }

        dat_state_ = UPDATED;
        auto dict = GetLocaleDictReference(locale);
        for (auto file : *dict) {
            if (dictionary_[file.first] == nullptr) {
                LOG(WARNING) << "In locale dictionary there is file with file_id = " << file.first
                             << "which is not in .dat file! Passing it and removing from locale dictionary";
                dict->erase(file.first);
                continue;
            }
            if (dictionary_[file.first]->MakeHeaderData().CutData(8, 16) ==
                file.second->MakeHeaderData().CutData(8, 16) || inactive_categories.count(orig_dict_[file.first]->category) != 0)
                continue;

            long long file_id = file.first;
            Subfile *new_file = file.second;

            dictionary_[file_id]->file_offset_ = new_file->file_offset_;
            dictionary_[file_id]->file_size_ = new_file->file_size_;
            dictionary_[file_id]->block_size_ = new_file->block_size_;
            dictionary_[file_id]->timestamp_ = new_file->timestamp_;
            dictionary_[file_id]->version_ = new_file->version_;

            patched_list.insert(file.first);
            dat_state_ = UPDATED;
        }

        current_locale_ = locale;
        CommitChanges();
        return SUCCESS;
    }

    bool DatFile::CheckIfUpdatedByGame() {
        if (!pending_patch_.empty())
            return true;
        if (current_locale_ == ORIGINAL)
            return false;

        bool updated = false;

        for (auto i : dictionary_) {
            long long file_id = i.first;
            Subfile* subfile = i.second;
            if (inactive_categories.count(subfile->category) > 0)
                continue;
            if (patch_dict_.count(file_id) > 0
                && (subfile->file_size() != patch_dict_[file_id]->file_size()
                || subfile->file_offset() != patch_dict_[file_id]->file_offset()
                || subfile->block_size() != patch_dict_[file_id]->block_size())) {
                orig_dict_.erase(file_id);
                patch_dict_.erase(file_id);
                pending_patch_.insert(file_id);
                updated = true;
                dat_state_ = UPDATED;
            }
        }
        CommitChanges();
        return updated;
    }

    DAT_RESULT DatFile::RepairPatches(Database *db) {
        // TODO: Find errors
        SubfileData data;
        data = db->GetNextFile();
        while (!data.Empty()) {
            if (pending_patch_.count(data.options["fid"].as<long long>()) > 0) {
                PatchFile(data);
            }
            data = db->GetNextFile();
        }
        CommitChanges();
        return SUCCESS;
    }

    DAT_RESULT DatFile::FinishRepairingPatches() {
        pending_patch_.clear();
        return SUCCESS;
    }

    LOCALE DatFile::current_locale() {
        if (dat_state_ < READY) {
            LOG(ERROR) << "dat_file is in incorrect state!";
            return ORIGINAL;
        }
        if (current_locale_ != PATCHED && current_locale_ != ORIGINAL) {
            LOG(ERROR) << "locale has incorrect value. Setting it to original";
            current_locale_ = ORIGINAL;
        }
        return current_locale_;
    }

    DAT_RESULT DatFile::CommitLocales() {
        std::cout << "Committing locales..." << std::endl;
        SubfileData data = dictionary_[2013266257]->PrepareForExport(GetFileData(dictionary_[2013266257]));
        data.options["fid"] = "2013266257";
        data.options["ext"] = ".unknown";

        BinaryData old_data = BinaryData(GetFileData(dictionary_[2013266257u]));

        // 14 bytes for old data
        // 15 bytes for "Hi from Gi1dor"
        // 4 bytes for LOCALE
        // 4 bytes for orig_dict.size()
        // (32 + 4) * orig_dict.size() bytes for orig_dict data
        // 4 bytes for patch_dict.size()
        // (32 + 4) * patch_dict.size() bytes for patch_dict data
        // 4 bytes for inactive_categories list
        // 4 * inactive_categories.size() bytes for inactive_categories data

        data.binary_data = BinaryData(14 + 15 + 4
                                      + 4 + (32 + 4) * orig_dict_.size()
                                      + 4 + (32 + 4) * patch_dict_.size()
                                      + 4 + 4 * inactive_categories.size());

        size_t current_size = 0;
        data.binary_data.Append(GetFileData(dictionary_[2013266257u]).CutData(0, 14), current_size);
        current_size += 14;

        data.binary_data.Append(BinaryData("Hi from Gi1dor!", 15), current_size);
        current_size += 15;

        data.binary_data.Append(BinaryData((current_locale_ == ORIGINAL ? "ORIG" : "PATC"), 4), current_size);
        current_size += 4;

        data.binary_data.Append(BinaryData::FromNumber<4>(orig_dict_.size()), current_size);
        current_size += 4;

        for (auto file : orig_dict_) {
            data.binary_data.Append(file.second->MakeHeaderData(), current_size);
            current_size += 32;
            data.binary_data.Append(BinaryData::FromNumber<4>(file.second->category), current_size);
            current_size += 4;
        }

        data.binary_data.Append(BinaryData::FromNumber<4>(patch_dict_.size()), current_size);
        current_size += 4;

        for (auto file : patch_dict_) {
            data.binary_data.Append(file.second->MakeHeaderData(), current_size);
            current_size += 32;
            data.binary_data.Append(BinaryData::FromNumber<4>(file.second->category), current_size);
            current_size += 4;
        }

        data.binary_data.Append(BinaryData::FromNumber<4>(inactive_categories.size()), current_size);
        current_size += 4;
        for (auto patch_id : inactive_categories) {
            data.binary_data.Append(BinaryData::FromNumber<4>(patch_id), current_size);
            current_size += 4;
        }

        DAT_RESULT result = PatchFile(data, true);
        if (result != SUCCESS)
            return result;

        return SUCCESS;
    }

    DAT_RESULT DatFile::EnableCategory(int category) {
        LOG(DEBUG) << "Enabling category " << category;
        if (inactive_categories.count(category) == 0)
            return SUCCESS;

        inactive_categories.erase(category);
        dat_state_ = UPDATED;

        for (auto file : dictionary_) {
            auto file_id = file.first;
            if (patch_dict_.count(file_id) > 0 && patch_dict_[file_id]->category == category) {
                file.second->file_offset_ = patch_dict_[file_id]->file_offset_;
                file.second->file_size_ = patch_dict_[file_id]->file_size_;
                file.second->block_size_ = patch_dict_[file_id]->block_size_;
                file.second->timestamp_ = patch_dict_[file_id]->timestamp_;
                file.second->version_ = patch_dict_[file_id]->version_;
                patched_list.insert(file_id);
            }
        }
        return SUCCESS;
    }

    DAT_RESULT DatFile::DisableCategory(int category) {
        LOG(DEBUG) << "Disabling category " << category;
        if (inactive_categories.count(category) != 0)
            return SUCCESS;
        inactive_categories.insert(category);
        dat_state_ = UPDATED;

        for (auto file : dictionary_) {
            auto file_id = file.first;
            if (orig_dict_.count(file_id) && orig_dict_[file_id]->category == category) {
                file.second->file_offset_ = orig_dict_[file_id]->file_offset_;
                file.second->file_size_ = orig_dict_[file_id]->file_size_;
                file.second->block_size_ = orig_dict_[file_id]->block_size_;
                file.second->timestamp_ = orig_dict_[file_id]->timestamp_;
                file.second->version_ = orig_dict_[file_id]->version_;
                patched_list.insert(file_id);
            }
        }
        return SUCCESS;
    }

    const std::set<long long>& DatFile::GetInactiveCategoriesList() {
        return inactive_categories;
    }

    bool DatFile::CheckIfNotPatched() {
        return patch_dict_.empty();
    }

    bool DatFile::CheckIfPatchedByOldLauncher() {
        return dictionary_.count(620750000) > 0;
    }

    const std::string &DatFile::filename() const {
        return filename_;
    }

}
}